
#include "SBCore.h"


void FSBlit()
{
	// the fullscreen triangle
	SUIVtx fst[] = 
	{
		{ mVec2( -3, -2 ), mVec2( -1, -0.5f ), 1.0f },
		{ mVec2( 0, 3 ), mVec2( 0.5f, 2 ), 1.0f },
		{ mVec2( 3, -2 ), mVec2( 2, -0.5f ), 1.0f },
	};
	DrawUIVtx( fst, 3, PT_Triangles );
}

FLOAT LayeredSineNoise( FLOAT x )
{
	FLOAT noise = 0;
	FLOAT periods[] = { 1.823f, 2.8934f, 3.2523f, 5.6342f, 7.32489f, 9.9832f };
	INT32 size = sizeof( periods ) / sizeof( FLOAT );
	for( INT32 i = 0; i < size; ++i )
		noise += gSin( x * periods[ i ] );
	noise /= size;
	return noise * 0.5f + 0.5f;
}


void SBMenu::Do()
{
	for( UINT32 i = 0; i < Options.Size(); ++i )
	{
		FLOAT size = 32.0f;
		mAABB2 rect( Pos, Pos + mVec2( 384, 32 ) );
		rect.lo.y += size * 1.5f * i;
		rect.hi.y += size * 1.5f * i;
		mFColor color( 0.6f, 1 );
		if( Pick == i )
		{
			color = mFColor( gMod( GetTransQ() * 7.9f, 1 ) > 0.5f ? 1.0f : 0.0f );
			FLOAT cn = LayeredSineNoise( Time + 891 ) * 0.9f;
			color.g -= cn;
			color.b -= cn;
			rect.lo.x -= 10;
			CurMaterial = &MatTexture;
			DrawTextRect( "|", 1, color, 0, rect, size );
			rect.lo.x += 20;
			rect.lo.y += -3 + LayeredSineNoise( Time * 0.4f + 123 ) * 6;
		}
		CurMaterial = &MatTexture;
		if( Visible && rect.Intersect( IO->Input.CursorPos ) )
		{
			Pick = i;
			if( IO->Input.KeyState( EK_LeftButton ) == 1 )
				Visible = FALSE;
		}
		DrawTextRect( Options[ i ].CStr(), Options[ i ].Size(), color, 0, rect, size );
	}

	if( Visible )
	{
		if( IO->Input.KeyState( EK_Up ) == KS_PRESSED ) Pick--;
		else if( IO->Input.KeyState( EK_Down ) == KS_PRESSED ) Pick++;
		Pick %= Options.Size();
		if( Pick < 0 ) Pick += Options.Size();
	}
	if( IO->Input.KeyState( EK_Space ) == KS_PRESSED || IO->Input.KeyState( EK_Return ) == KS_PRESSED )
	{
		Visible = FALSE;
	}
}


void PFX_GenFloats( FLOAT* data, UINT32 sz, FLOAT fmin, FLOAT fmax )
{
	FLOAT* dend = data + sz;
	while( data != dend )
		*data++ = Lerp( fmin, fmax, M_RandF() );
}

void PFX_GenVelAD( mVec2* data, UINT32 sz, mVec2 dir, FLOAT dvg, FLOAT lmin, FLOAT lmax )
{
	mVec2* dend = data + sz;
	while( data != dend )
	{
		FLOAT len = Lerp( lmin, lmax, M_RandF() );
		FLOAT ang = ( M_RandF() * 2 - 1 ) * dvg * PI;
		*data++ = dir.Rotated( ang ) * len;
	}
}

UBOOL PFX_AnyAbove( const FLOAT* data, UINT32 sz, FLOAT val )
{
	const FLOAT* dend = data + sz;
	while( data != dend )
		if( *data++ > val )
			return TRUE;
	return FALSE;
}



void SBSparkFX::Trigger( mVec2 pos, mVec2 dir, FLOAT mult )
{
	Pos = pos;
	PFX_GenVelAD( Vel, PartCount, dir, 0.2f, 2.0f * mult, 3.0f * mult );
	PFX_GenFloats( Alpha, PartCount, 0.8f, 1.1f );
	Time = 0;
	State = TRUE;
}

void SBSparkFX::Advance( FLOAT delta )
{
	Time += delta * 5;
	State = PFX_AnyAbove( Alpha, PartCount, Time );
}

void SBSparkFX::Render( const mMatrix& mtx )
{
	if( !State )
		return;

	SUIVtx vts[ PartCount * 2 ];
	for( UINT32 i = 0; i < PartCount; ++i )
	{
		vts[ i * 2 ].Pos = Pos + Vel[ i ] * Time;
		vts[ i * 2 ].Tex = mVec2( 0.0f );
		vts[ i * 2 ].Col = mFColor( 2.1f, 1.3f, 0.4f, Alpha[ i ] - Time );
		vts[ i * 2 + 1 ] = vts[ i * 2 ];
		vts[ i * 2 ].Pos += Vel[ i ] * 0.1f;
		vts[ i * 2 + 1 ].Col.a -= 0.5f;
	}
	M_SetShading( MBM_ADD, Shader );
	Shader->GetVariable( "SG_mProj" )->SetF( (FLOAT*) mtx.fa, 4, TRUE );
	DrawUIVtx( vts, PartCount * 2, PT_Lines );
}


SBExplodeFX::SBExplodeFX() : Time( 1 )
{
	Shader = SH_Get( "Shaders/Flash.shd" );
	Star.Load( "Textures/ExpStar.png", TEX_NOMIPMAPS | TEX_NOSRGB );
	Flame.Load( "Textures/ExpFlame.png", TEX_NOMIPMAPS | TEX_NOSRGB );
	GSound.SetupEmitter( "Sound/sb_explode.wav", &Sound );
}

void SBExplodeFX::Trigger( mVec2 pos, FLOAT size )
{
	Time = 0;
	Pos = pos;
	Size = size;
	for( UINT32 i = 0; i < StarCount; ++i )
	{
		StarAngles[ i ] = M_RandF() * PI * 2;
		StarAVel[ i ] = ( M_RandF() * 2 - 1 );
	}
	for( UINT32 i = 0; i < FlameCount; ++i )
	{
		FlameAngles[ i ] = M_RandF() * PI * 2;
		FlameAVel[ i ] = ( M_RandF() * 2 - 1 );
	}
	Sound.Stop();
	Sound.SetDistanceFactor( 5.0f );
	Sound.SetVolume( 0.75f );
	Sound.SetPosition( mVec3( Pos ) );
	Sound.Play();
}

void SBExplodeFX::DrawSprite( FLOAT angle, FLOAT size, FLOAT alpha )
{
	mVec2 dn = -mVec2( gSin( angle ), gCos( angle ) ) * size;
	mVec2 rt = dn.Perp2();
	mFColor color( 1, alpha );
	SUIVtx vts[] =
	{
		{ Pos - dn - rt, mVec2( 0, 0 ), color },
		{ Pos - dn + rt, mVec2( 1, 0 ), color },
		{ Pos + dn + rt, mVec2( 1, 1 ), color },
		{ Pos + dn - rt, mVec2( 0, 1 ), color },
	};
	DrawUIVtx( vts, 4, PT_TriangleFan );
}

void SBExplodeFX::Advance( FLOAT delta )
{
	Time += delta;
	if( Time < 1 )
	{
		for( UINT32 i = 0; i < StarCount; ++i )
			StarAngles[ i ] += StarAVel[ i ] * delta;
		for( UINT32 i = 0; i < FlameCount; ++i )
			FlameAngles[ i ] += FlameAVel[ i ] * delta;
	}
}

void SBExplodeFX::Render( const mMatrix& view )
{
	if( Time >= 1 )
		return;

	M_SetShading( MBM_ADD, Shader );
	Shader->GetVariable( "SG_mProj" )->SetF( (FLOAT*) view.fa, 4, TRUE );

	M_SetTexture( 0, Flame );
	for( UINT32 i = 0; i < FlameCount; ++i )
		DrawSprite( FlameAngles[ i ], Size * ( 1 + Time ), 1 - Time );
	
	M_SetTexture( 0, Star );
	for( UINT32 i = 0; i < StarCount; ++i )
		DrawSprite( StarAngles[ i ], gPow( ( 1 - Time ), 4.0f ) * Size * 2, gPow( 1 - Time, 4.0f ) );
}

