

#pragma once

#include "SBCore.h"


struct SBPhyBody
{
	SBPhyBody() : Pos( 0.0f ), Vel( 0.0f ), 
		Angle( 0.0f ), AVel( 0.0f ), CrashAmount( 0.0f ){}

	mVec2 Pos, Vel;
	FLOAT Angle, AVel;
	FLOAT CrashAmount;
};

//
// P l a y e r
//

struct SBPlayer : SBPhyBody
{
	SBPlayer(){ Reset(); }

	// combat
	FLOAT ShootTimeout;	// time to weapon unblocking
	INT32 Ammo;
	INT32 Health;

	// misc
	FLOAT CrashAmount;	// amount of camera shaking
	UBOOL Exists;		// no players exist in the menu level
	FLOAT LifeState;	// intro/outro (spawn/die)

	SBSparkFX FxSparks;

	void Reset()
	{
		// Pos is set in the initialization code, along with 'Exists'
		Vel.Set( 0, 0 );
		Angle = 0;
		AVel = 0;
		ShootTimeout = 0;
		Ammo = 100;
		Health = 20;
		CrashAmount = 0;
		Exists = FALSE;
		LifeState = -1;
	}
};

//
// E n e m y  [AI]
//
/*

States:
 - patrol
 - sweep

*/

struct SBEnemy : SBPhyBody
{
	SBEnemy() : ShootTimeout( 2.0f ), Health( 4 ),
		LifeState( -1 ), AlertState( 0 ), MoveDir( 0, 1 ), 
		LastKnownPos( 0.0f ), AngleTgt( 0 ), TmSideCheck( 0 ), 
		TmRoundCheck( 0 ), TmChkDecide( 0 ), ViewTime( 0 )
	{
		GSound.SetupEmitter( "Sound/sb_shoot.wav", &SndShoot );
		SndShoot.Set3DMode( TRUE );
		SndShoot.SetVolume( 0.95f );
		SndShoot.SetDistanceFactor( 500.0f );
	}

	// combat
	FLOAT ShootTimeout;
	INT32 Health;

	// AI
	// - general
	mVec2 OrigPos;
	FLOAT LifeState;	// intro/outro (spawn/die)
	FLOAT AlertState;	// alert timeout
	// - movement
	mVec2 MoveDir;		// direction of movement
	mVec2 LastKnownPos;	// last position of the player character
	// - watching
	FLOAT AngleTgt;		// target angle
	FLOAT TmSideCheck;	// side check time(out) [sequence playback position]
	FLOAT TmRoundCheck;	// turn-around check time(out) [^spp]
	FLOAT TmChkDecide;	// time(out) before next check decision
	FLOAT ViewTime;		// time of looking at the player
	SSEmitter SndShoot;	// shooting sound
	SBExplodeFX DieFX;	// death explosion FX
};
typedef TArray< SBEnemy > EnemyVec;

//
// B u l l e t
//

struct SBBullet
{
	mVec2 Pos;
	mVec2 Dir;		// normalized
	FLOAT Timeleft;
	FLOAT Lifetime;
	FLOAT Distance;	// max distance
	FLOAT RayDist;	// distance to hit
	UBOOL HurtPlayer;	// target
	FLOAT Random;	// for FX consistency

	SBSparkFX HitFX;	// hit effects
};
typedef TArray< SBBullet > BltVec;

//
// C o l l e c t i b l e
//

struct SBCollect
{
	mVec2 Pos;
	FLOAT TimeState;
};
typedef TArray< SBCollect > ClcVec;


