

#include "SGFileIO.h"

#include "SBGame.h"


#define TRANSTIME 2


SBGame::SBGame( SGClientWindow* win ) : Paused( TRUE ), InGame( FALSE ), NeedToQuit( FALSE ), 
	LevelTime( 0 ), LevelStopTime( FLT_MAX ), EndActType( ENDACT_LOADING ), LevelID( 1 )
{
	AutoCleanup = FALSE;
	IO = win;
	UpdatePlacement();

	Level = new SBLevel( win );
	Level->Load( "Levels/menu.txt" );
	TexNoise.Load( "Textures/VNoise.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );

	MainMenu = new SBMenu;
	MainMenu->Pos = mVec2( 60, 400 );
	MainMenu->Options.PushBack( "Play" );
	MainMenu->Options.PushBack( "Quit" );
	AddLayer( MainMenu );
	MainMenu->UpdatePlacement();

	GSound.SetupEmitter( "Music/killstealer.ogg", &SndMusic );
	SndMusic.SetVolume( 0.65f );
	SndMusic.SetLooping( TRUE );
}

SBGame::~SBGame()
{
	delete Level;
}

void SBGame::Advance( FLOAT delta )
{
	if( Paused )
		Level->Process( delta );
	else
		Level->Advance( delta );

	if( InGame && IO->Input.KeyState( EK_Escape ) == 1 )
	{
		Paused = !Paused;
		MainMenu->Visible = Paused;
		MainMenu->Transition = !Paused;
	}

	Process( delta );

	if( !MainMenu->Visible && MainMenu->Transition == 0 )
	{
		if( MainMenu->Pick != 0 || MainMenu->Options[ 0 ] != "Resume" )
		{
			LevelStopTime = LevelTime + TRANSTIME - 0.001f;
			EndActType = ( MainMenu->Pick == 1 && MainMenu->Options[ 1 ] == "Quit" ) ? ENDACT_QUITTING : ENDACT_LOADING;
			if( MainMenu->Pick == 1 )
				SndMusic.FadeVolume( 0, 2 );
		}
		else
		{
			LevelStopTime = LevelTime - 1;
		}
	}

	// End action triggers
	if( InGame && EndActType == ENDACT_LOADING && Level->Collectibles.Size() == 0 )
	{
		EndActType = ENDACT_SUCCESS;
		LevelStopTime = LevelTime + TRANSTIME;
	}
	if( InGame && EndActType == ENDACT_LOADING && Level->Player.LifeState == 0 && Level->Timeout <= 0 )
	{
		EndActType = ENDACT_RESTART_TIMEOUT;
		LevelStopTime = LevelTime + TRANSTIME;
	}
	if( InGame && EndActType == ENDACT_LOADING && Level->Player.LifeState > 0 )
	{
		EndActType = ENDACT_RESTART_DEATH;
		LevelStopTime = LevelTime + TRANSTIME;
	}
	if( IO->Input.KeyState( EK_LControl ) == 2 && IO->Input.KeyState( EK_R ) == 1 )
	{
		EndActType = ENDACT_RESTART_MANUAL;
		LevelStopTime = LevelTime + TRANSTIME;
	}

	LevelTime += delta;
	FLOAT ltq1 = MAX( MIN( ( LevelStopTime - LevelTime ) / TRANSTIME, 1 ), 0 );
	FLOAT ltq2 = MAX( MIN( ( LevelTime ) / TRANSTIME, 1 ), 0 );
	FLOAT ltq = ltq1 * ltq2;
	MainMenu->Time += delta;
	
	if( ltq > 0 )
	{
		M_SetShading( MBMF_BLEND( MBMF_BLEND_SUB, MBMF_BF_SA, MBMF_BF_ONE ), SH_Get( "Shaders/FSLineFade.shd" ) );
		FLOAT qval = 1 - MIN( 1, ltq );
		M_GetCurShader()->GetVariable( "Q" )->SetF( &qval, 1 );
		M_SetTexture( 0, TexNoise );
		FSBlit();
		if( ltq1 < 1 || EndActType == ENDACT_ENDING )
		{
			CurMaterial = &MatTexture;
			const TCHAR* Endings[] = 
			{
				"Success...",
				"No more time!",
				"You died!",
				"Restarting...",
				"loading...",
				"\n\n\n\n\nlet's kill some bots\nagain soon!",
				"You rock!\nAll levels are completed!\n\nthanks for playing",
			};
			const TCHAR* out = Endings[ EndActType ];
			DrawTextRect( out, 0, mFColor( 0.7f, 0.8f, 0.9f, 1 - ltq1 + ( EndActType == ENDACT_ENDING ) ), TF_CENTER | TF_VCENTER, Rect, 64.0f );
		}
		if( ltq2 < 1 && InGame )
		{
			if( EndActType == ENDACT_TOENDING )
			{
				EndActType = ENDACT_ENDING;
				Level->DrawUI = FALSE;
				SndMusic.FadeVolume( 0, 3 );
			}
			else if( EndActType != ENDACT_ENDING )
				EndActType = ENDACT_LOADING;
			CurMaterial = &MatTexture;
			TString filename;
			filename.Printf( 32, "Level %03d", LevelID );
			DrawTextRect( filename.CStr(), filename.Size(), mFColor( 0.7f, 0.8f, 0.9f, 1 - ltq2 ), TF_CENTER | TF_VCENTER, Rect, 64.0f );
		}
	}
	else if( !MainMenu->Visible && Paused )
	{
		if( InGame )
		{
			if( MainMenu->Pick == 0 )
			{
				Paused = FALSE;
				MainMenu->Visible = Paused;
				MainMenu->Transition = 1;
			}
			else
			{
back_to_menu:
				if( SndMusic.IsPlaying() )
					SndMusic.Stop();
				Level->Load( "Levels/menu.txt" );
				LevelTime = 0;
				LevelStopTime = FLT_MAX;
				Paused = TRUE;
				InGame = FALSE;
				MainMenu->Options[ 0 ] = "Play";
				MainMenu->Options[ 1 ] = "Quit";
				MainMenu->TransTimes = 1.0f;
				MainMenu->Transition = -0.999f;
				MainMenu->Visible = 1;
				Level->DrawUI = FALSE;
			}
		}
		else
		{
			// execute..
			if( MainMenu->Pick == 0 )
			{
				// Play
				SndMusic.SetVolume( 1 );
				if( SndMusic.IsPlaying() )
					SndMusic.Stop();
				SndMusic.Play();
				LevelID = 1;
				LoadLevelByID();
				Paused = FALSE;
				InGame = TRUE;
				MainMenu->Options[ 0 ] = "Resume";
				MainMenu->Options[ 1 ] = "Quit to menu";
				MainMenu->TransTimes = 0.01f;
				Level->DrawUI = TRUE;
			}
			else
			{
				// Quit
				NeedToQuit = TRUE;
			}
		}
	}

	if( LevelTime > LevelStopTime )
	{
		if( EndActType == 0 )
		{
			LevelID++;
			if( HasLevel() )
				LoadLevelByID();
			else
			{
				// load credits
				LevelID = 1337;
				Level->Load( "Levels/ending.png" );
				Level->CamPos.Set( 15, 7 );
				Level->CamSize = 10;
				LevelTime = 0;
				LevelStopTime = 10;
				EndActType = ENDACT_TOENDING;
			}
		}
		else if( EndActType >= 1 && EndActType <= 3 )
		{
			LoadLevelByID();
		}
		else if( EndActType == ENDACT_ENDING )
		{
			EndActType = ENDACT_LOADING;
			goto back_to_menu;
		}
		else if( EndActType == ENDACT_TOENDING )
			EndActType = ENDACT_ENDING;
	}
}

void SBGame::Do()
{
	if( Paused || !InGame )
	{
		// background
		CurMaterial = &MatColor;
		DrawColRect( Rect, mFColor( 0, 0.5f ), FALSE );

		// UI
		const TCHAR* TITLE = InGame ? "--= PAUSED =--" : "--= Silver Bullet =--";
		CurMaterial = &MatTexture;
		mAABB2 logorect( Rect.lo.x, Rect.lo.y, Rect.hi.x, Rect.Center().y );
		mFColor textcol( 0.6f, 0.605f, 0.61f );
		DrawTextRect( TITLE, 0, textcol * mFColor( 0.1f, 0.5f ), TF_CENTER | TF_VCENTER, logorect, 52.0f );
		DrawTextRect( TITLE, 0, textcol, TF_CENTER | TF_VCENTER, logorect, 48.0f );
		logorect.lo += mVec2( 4.0f );
		DrawTextRect( TITLE, 0, ( textcol * 0.4f ).Opaque() * 0.5f, TF_CENTER | TF_VCENTER, logorect, 48.0f );

		if( !InGame )
		{
			mVec2 SB( 64.0f ), SB2( 66.0f ), SB0( 32.0f );
			const TCHAR* INSTR = "\
								 ***MISSION***\n\
								 collect all shiny orbs\n\
								 avoid or kill the angry bots\n\n\
								 ***CONTROLS***\n\
								 [W/A/S/D] - move\n\
								 [mouse] - aim\n\
								 [LMB] - shoot\n\n\
								 ***CREDITS***\n\
								 game code|graphics|sfx - snake5\n\
								 music by Waterflame (Newgrounds)\n\
								 ";
			DrawTextRect( INSTR, 0, mFColor( 0.1f, 0.06f, 0.03f, 0.9f ), TF_RIGHT | TF_BOTTOM, mAABB2( Rect.lo + SB0, Rect.hi - SB2 ), 32.0f );
			DrawTextRect( INSTR, 0, mFColor( 0.8f, 0.88f, 0.96f, 0.9f ), TF_RIGHT | TF_BOTTOM, mAABB2( Rect.lo + SB0, Rect.hi - SB ), 32.0f );
		}
	}
}

TString SBGame::GetLevelFileName()
{
	TString file;
	file.Printf( 128, "Levels/level%03d.png", LevelID );
	return file;
}

UBOOL SBGame::HasLevel()
{
	return F_FileExists( GetLevelFileName() );
}

void SBGame::LoadLevelByID()
{
	TString file = GetLevelFileName();
	BREAK_IF( !HasLevel() );
	Level->Load( file.CStr() );
	LevelTime = 0;
	LevelStopTime = FLT_MAX;
	EndActType = ENDACT_LOADING;
}
