

#include "SGImage.h"
#include "SGFileIO.h"

#include "SBLevel.h"



const FLOAT PLAYER_RADIUS = 0.35f;
const FLOAT ENEMY_RADIUS = 0.35f;
const FLOAT COLLECT_COL_RADIUS = 0.35f;//0.125f;
const FLOAT COLLECT_RADIUS = 0.5f;//0.25f;
const UINT32 AMMO_DIV = 10;



SBLevel::SBLevel( SGClientWindow* win ) : CamPos( 7.0f, 6.0f ), CamSize( 5.0f ), DrawUI( FALSE )
{
	IO = win;
	UpdatePlacement();

	TexPlayer.Load( "Textures/Player.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexEnemy.Load( "Textures/Enemy.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexWall.Load( "Textures/Wall.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexFloor.Load( "Textures/Floor.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexFlash.Load( "Textures/Flash.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexCollect.Load( "Textures/Collect.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexColNoise.Load( "Textures/ColNoise.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	TexEnemyLight.Load( "Textures/EnemyLight.png", TEX_NOSRGB | TEX_NOMIPMAPS | TEX_NOSTREAM );
	COLOR32 cnull = 0;
	TexNull.Create( 1, 1, 1, 1, TEXFMT_RGBA8, &cnull, TEX_NOMIPMAPS );
	// calculate overlay textures
	// bits: -x, +x, -y, +y
	UINT32 size = 16;
	SImage2D img( size, size );
	for( UINT32 i = 0; i < 16; ++i )
	{
		for( UINT32 p = 0; p < img.Data.Size(); ++p )
			img.Data[ p ] = 0xff555555;
		if( ( i & 1 ) == 0 ){ for( UINT32 p = 0; p < img.Height; ++p ) img.Get( 0, p ) = 0xff111111; }
		if( ( i & 2 ) == 0 ){ for( UINT32 p = 0; p < img.Height; ++p ) img.Get( img.Width - 1, p ) = 0xff111111; }
		if( ( i & 4 ) == 0 ){ for( UINT32 p = 0; p < img.Width; ++p ) img.Get( p, 0 ) = 0xff111111; }
		if( ( i & 8 ) == 0 ){ for( UINT32 p = 0; p < img.Width; ++p ) img.Get( p, img.Height - 1 ) = 0xff111111; }
		TexWallOvr[ i ].Create( img.Width, img.Height, 1, 1, TEXFMT_RGBA8, img.GetBits(), TEX_NOMIPMAPS );
	}
	TexLighting.Create( IO->GetWidth(), IO->GetHeight(), 1, 1, TEXFMT_RGBA16F, NULL, TEX_NOMIPMAPS );
	FBLighting = new SFrameBuffer;
	FBLighting->SetTex( 0, TexLighting, FB_COLOR ).Finish();
	TexRB.Create( IO->GetWidth(), IO->GetHeight(), 1, 1, TEXFMT_RGBA16F, NULL, TEX_NOMIPMAPS );
	FBRB = new SFrameBuffer;
	FBRB->SetTex( 0, TexRB, FB_COLOR ).Finish();

	GSound.SetupEmitter( "Sound/sb_move.wav", &SndPlayerMove );
	GSound.SetupEmitter( "Sound/sb_turn.wav", &SndPlayerTurn );
	GSound.SetupEmitter( "Sound/sb_shoot.wav", &SndPlayerShoot );
	SndPlayerTurn.SetVolume( 0.5f );
	SndPlayerMove.SetVolume( 0.0f );
	SndPlayerMove.SetLooping( TRUE );
	SndPlayerMove.Play();
	SndPlayerShoot.Set3DMode( TRUE );
	SndPlayerShoot.SetVolume( 0.75f );
	SndPlayerShoot.SetDistanceFactor( 5.0f );
	GSound.SetupEmitter( "Sound/sb_pickup.wav", &SndOrbHit );
	SndOrbHit.SetVolume( 0.75f );
}

void SBLevel::LoadFromText( const TCHAR* txt )
{
	Unload();

	UINT32 x = 0, y = 0, w = 0, h = 0;
	// Pass 1 - measurement
	const TCHAR* otxt = txt;
	while( *txt )
	{
		if( *txt == '\n' )
		{
			w = MAX( x, w );
			y++;
			x = 0;
		}
		else
			x++;
		txt++;
	}
	h = MAX( y + 1, h );
	Walls.Resize( w * h );
	for( UINT32 i = 0; i < Walls.Size(); ++i )
		Walls[ i ] = 0;
	Width = w;
	Height = h;
	txt = otxt;

	SBEnemy enm;
	SBCollect clc = { 0.0f, 1.0f };

	// Pass 2 - loading
	x = 0; y = 0;
	while( *txt )
	{
		switch( *txt )
		{
		case '#': // wall
			Walls[ x + y * w ] = 1;
			break;
		case 'P': // player
			Player.Exists = TRUE;
			Player.Pos = mVec2( x, y );
			break;
		case 'L': // light
			Lights.PushBack( mVec2( x, y ) );
			break;
		case 'E': // enemy
			enm.Pos.Set( x, y );
			enm.OrigPos = enm.Pos;
			Enemies.PushBack( enm );
			break;
		case '*': // collectible
			clc.Pos.Set( x, y );
			Collectibles.PushBack( clc );
			break;
		case '\n': // new line
			x = 0;
			y++;
			goto noadvance;
			break;
		default: // empty space
			break;
		}
		x++;
noadvance:
		txt++;
	}

	// Pass 3 - calculate adjacencies
	Adjacency.Resize( Walls.Size() );
	for( UINT32 i = 0; i < Walls.Size(); ++i )
	{
		INT32 x = i % Width;
		INT32 y = i / Width;
		UINT8 adj = 0;
		if( x > 0 && Walls[ x - 1 + y * Width ] ) adj |= 1;
		if( x < Width - 1 && Walls[ x + 1 + y * Width ] ) adj |= 2;
		if( y > 0 && Walls[ x + ( y - 1 ) * Width ] ) adj |= 4;
		if( y < Height - 1 && Walls[ x + ( y + 1 ) * Width ] ) adj |= 8;
		Adjacency[ i ] = adj;
	}

	// Pass 4 - lightmap calculation
	SImage2D img;
	UINT32 mult = 4;
	img.Resize( Width * mult, Height * mult );
	for( UINT32 py = 0; py < img.Height; ++py )
	{
		for( UINT32 px = 0; px < img.Width; ++px )
		{
			UINT32 lx = px / mult;
			UINT32 ly = py / mult;
			if( Walls[ lx + ly * Width ] )
				img.Get( px, py ) = 0xff000000;
			else
			{
				const UINT32 qtbl[] = {
					63, 21, 7, 1, 
					1, 7, 21, 63 };
				UINT32 qx = px - lx * mult;
				UINT32 qy = py - ly * mult;
				UINT8 q = 255;
				if( Adjacency[ lx + ly * Width ] & 1 ) q -= qtbl[ qx ];
				if( Adjacency[ lx + ly * Width ] & 2 ) q -= qtbl[ qx + 4 ];
				if( Adjacency[ lx + ly * Width ] & 4 ) q -= qtbl[ qy ];
				if( Adjacency[ lx + ly * Width ] & 8 ) q -= qtbl[ qy + 4 ];
				COLOR32 col = COLOR_RGBX( q, q, q );
				img.Get( px, py ) = col;
			}
		}
	}
	TexLightmap.Create( img.Width, img.Height, 1, 1, TEXFMT_RGBA8, img.GetBits(), TEX_NOMIPMAPS );

	// the rest
	Time = M_RandF();
}

void SBLevel::Load( const TCHAR* fname )
{
	TString out;
	if( Path_GetExt( fname ).ToLower() == "png" )
	{
		SImage2D img;
		IMG_LoadF( fname, &img );
		out.Reserve( img.Data.Size() + img.Height );
		for( UINT32 i = 0; i < img.Data.Size(); ++i )
		{
			CHAR c = ' ';
			switch( img.Data[ i ] )
			{
			case 0xff000000: c = '#'; break;
			case 0xffff0000: c = 'P'; break;
			case 0xffff00ff: c = 'L'; break;
			case 0xff0000ff: c = 'E'; break;
			case 0xff00ffff: c = '*'; break;
			}
			out.PushBack( c );
			if( ( i + 1 ) % img.Width == 0 && i + 1 != img.Data.Size() )
				out.PushBack( '\n' );
		}
	}
	else
	{
		// interpret as text
		ReadTextFileToString( fname, out );
	}
	LoadFromText( out.CStr() );

	TString fn2 = fname;
	fn2 += ".txt";
	TString data;
	if( ReadTextFileToString( fn2.CStr(), data ) )
	{
		INT32 at = 0;
		while( at < (INT32) data.Size() )
		{
			F_SkipWhitespaces( data, at );
			TString key = F_ReadString( data, at, &FStr_Token );
			key = key.ToLower();
			if( key == "timeout" )	Timeout = F_ReadFloat( data, at );
		}
	}
}

void SBLevel::Unload()
{
	Player.Reset();
	Walls.Clear();
	Adjacency.Clear();
	Lights.Clear();
	Bullets.Clear();
	Collectibles.Clear();
	Enemies.Clear();
	CamPos.Set( 7, 6 );
	CamSize = 5;
	Timeout = 600;
}



/// - - - -
//
//   G A M E   L O G I C
//
/// - - - -

void SBLevel::Advance( FLOAT delta )
{
	Time += delta;

	// Process the player
	if( Player.Exists )
	{
		// ticking
		if( Player.LifeState < 0 )
			Player.LifeState = MIN( 0, Player.LifeState + delta );
		else if( Player.LifeState > 0 )
		{
			Player.LifeState = MIN( 1, Player.LifeState + delta );
		}
		else
		{
			// movement
			Player.Vel *= 0.97f;

			mVec2 tgtvel( 0.0f );
			if( IO->Input.KeyState( EK_A ) ) tgtvel.x -= 1;
			if( IO->Input.KeyState( EK_D ) ) tgtvel.x += 1;
			if( IO->Input.KeyState( EK_W ) ) tgtvel.y -= 1;
			if( IO->Input.KeyState( EK_S ) ) tgtvel.y += 1;
			if( tgtvel.LengthSq() )
			{
				tgtvel.Normalize();
				Player.Vel = LerpV2( Player.Vel, tgtvel * 12.0f, 0.07f );
			}
		}

		FLOAT aspect = FLOAT( IO->GetWidth() ) / FLOAT( IO->GetHeight() );
		mVec2 worldcp = ( IO->Input.CursorPos / mVec2( IO->GetWidth(), IO->GetHeight() ) * 2 - mVec2(1.0f) ).Mul( mVec2( aspect, 1 ) ) * CamSize + CamPos;

		FLOAT newang = gAtan2( worldcp.x - Player.Pos.x, worldcp.y - Player.Pos.y );
		if( newang - Player.Angle > PI ) newang -= PI * 2;
		else if( Player.Angle - newang > PI ) newang += PI * 2;
		Player.AVel = newang - Player.Angle;
		Player.AVel *= 6.0f;

		Player.Pos += Player.Vel * delta;

		Player.CrashAmount = MAX( Player.CrashAmount - delta * 10.0f, 0 );
		Player.CrashAmount *= 0.9f;

		Player.Angle += Player.AVel * delta;
		Player.Angle = gMod( Player.Angle, PI * 2 );
		CamPos = Player.Pos + mVec2( LayeredSineNoise( Time * 17.892f + 8.89203f ), LayeredSineNoise( Time * 12.8293f + 23.5223f ) ) * 
			( Player.ShootTimeout * 0.5f + Player.CrashAmount * 0.25f ) * 2 + Player.Vel * 0.01f;

		SndPlayerMove.SetVolume( Player.Vel.Length() * 0.3f );

		// shooting
		if( Player.ShootTimeout <= 0 && Player.Ammo > 0 && Player.LifeState == 0 )
		{
			if( IO->Input.KeyState( EK_LeftButton ) == KS_HOLDING )
			{
				Player.ShootTimeout = 0.10f;
				mVec2 spos = Player.Pos;
				FLOAT sang = Player.Angle;
				mVec2 sdir( gSin( sang ), gCos( sang ) );
				Shoot( spos, sdir, 10.0f, 0.1f, 0.1f, Enemies.Size() );
				Player.Vel -= sdir;
				Player.Ammo --;
				SndPlayerShoot.Stop();
				SndPlayerShoot.SetPosition( Player.Pos );
				SndPlayerShoot.Play();
				Player.FxSparks.Trigger( spos, sdir );
			}
		}
		else
			Player.ShootTimeout = MAX( 0, Player.ShootTimeout - delta );

		Player.FxSparks.Advance( delta );
	}
	else SndPlayerMove.SetVolume( 0 );

	GSound.SetListenerPosition( mVec3( CamPos.x, CamPos.y, -5 ) );
	GSound.SetListenerDirection( mVec3( 0, 0, 1 ), mVec3( 0, -1, 0 ) );

	// Process the enemies
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		SBEnemy& E = Enemies[ i ];
		E.DieFX.Advance( delta );

		if( E.LifeState < 0 )
			E.LifeState = MIN( 0, E.LifeState + delta );
		else if( E.LifeState > 0 )
			E.LifeState = MIN( 5, E.LifeState + delta );

		if( E.LifeState != 0 )
		{
			// respawn after 5
			if( E.LifeState >= 5 )
			{
				E.Health = 4;
				E.AlertState = 0;
				E.Pos = Enemies[ gRand() % Enemies.Size() ].OrigPos;
				E.LifeState = -1;
			}
			goto enemy_crashed;
		}

		// misc actions
		// - crash recovery
		if( E.CrashAmount > 0 )
		{
			E.CrashAmount -= delta;
enemy_crashed:
			// just a bit of it
			E.Angle += E.AVel * delta;
			E.Angle = gMod( E.Angle, PI * 2 );
			E.Pos += E.Vel * delta;
			continue;
		}

		// - checks
		FLOAT alsq = E.AlertState > 0;
		if( E.TmRoundCheck > 0 )
			E.TmRoundCheck -= delta * 2.5f + alsq;
		if( E.TmSideCheck > 0 )
			E.TmSideCheck -= delta * 2.5f + alsq;
		if( E.TmChkDecide > 0 )
			E.TmChkDecide -= delta;
		else
		{
			E.TmChkDecide = E.AlertState > 0 ? M_RandF() * 1 + 2 : M_RandF() * 2 + 3;
			if( M_RandF() > 0.5f )
				E.TmSideCheck = PI * 2;
			else
				E.TmRoundCheck = PI * 2;
		}


		// alert-specific actions
		if( E.ShootTimeout > 0 ) E.ShootTimeout -= delta;

		if( E.AlertState > 0 )
		{
			E.AlertState -= delta;
			E.ViewTime -= delta;

			mVec2 shdir = E.LastKnownPos - E.Pos;
			FLOAT shlen = shdir.Length();
			if( shlen )
				shdir /= shlen;

			if( E.ShootTimeout <= 0 )
			{
				Shoot( E.Pos, shdir, 10, 0.1f, 0.2f, i );
				E.SndShoot.SetDistanceFactor( 5.0f );
				E.SndShoot.SetVolume( 0.75f );
				E.SndShoot.Stop();
				E.SndShoot.SetPosition( E.Pos );
				E.SndShoot.Play();
				E.ShootTimeout = 0.1f;
			}
		}
		
		mVec2 endir( gSin( Enemies[ i ].Angle ), gCos( Enemies[ i ].Angle ) );
		mVec2 rcdir = Player.Pos - E.Pos;
		FLOAT rclen = rcdir.Length();
		if( rclen )
			rcdir /= rclen;
		if( rclen < 8.0f && ( rcdir * endir ) > 0.7f && RaycastWalls( E.Pos, rcdir, rclen ) == rclen )
		{
			if( RaycastEnts( E.Pos, rcdir, rclen, i ) == Enemies.Size() )
			{
				E.AlertState = 5;
				E.ViewTime = 1;
				E.LastKnownPos = Player.Pos;
			}
		}


		// processing
		// - angle physics (targeting)
		E.AngleTgt = E.AlertState > 0 ? gAtan2( E.LastKnownPos.x - E.Pos.x, E.LastKnownPos.y - E.Pos.y ) : gAtan2( E.MoveDir.x, E.MoveDir.y );
		E.AngleTgt += gSin( E.TmSideCheck ) * PI * 0.47f + E.TmRoundCheck;
		if( E.AngleTgt - E.Angle > PI ) E.AngleTgt -= PI * 2;
		else if( E.Angle - E.AngleTgt > PI ) E.AngleTgt += PI * 2;
		E.AVel = E.AngleTgt - E.Angle;
		E.AVel *= 5.0f;
		E.Angle += E.AVel * delta;
		E.Angle = gMod( E.Angle, PI * 2 );

		// - position physics (movement)
		bool rcres = RaycastWalls( E.Pos, E.MoveDir, 1 ) >= 1;
		if( ( E.AlertState <= 0 || ( E.Pos - Player.Pos ).Length() > 2 ) && rcres )
		{
			E.Vel = Lerp( E.Vel, E.MoveDir * ( 3 + alsq ), 0.1f );
			FLOAT rand = M_RandF();
			const FLOAT rcon = 0.01f;
			if( rand < rcon )
			{
				if( rand < rcon * 0.5f )
				{
					if( RaycastWalls( E.Pos, E.MoveDir.Perp(), 1 ) >= 1 )
						E.MoveDir = E.MoveDir.Perp();
				}
				else
				{
					if( RaycastWalls( E.Pos, E.MoveDir.Perp2(), 1 ) >= 1 )
						E.MoveDir = E.MoveDir.Perp2();
				}
			}
		}
		else
		{
			E.Vel *= 0.95f;
			if( !rcres )
			{
				E.Vel *= 0.7f;
				// pick a new direction
				if( M_RandF() > 0.5f )
					E.MoveDir = E.MoveDir.Perp();
				else
					E.MoveDir = E.MoveDir.Perp2();
			}
		}
		E.Pos += E.Vel * delta;
	}

	// Process the bullets
	for( UINT32 i = 0; i < Bullets.Size(); ++i )
	{
		Bullets[ i ].HitFX.Advance( delta );
		Bullets[ i ].Timeleft -= delta;
		if( Bullets[ i ].Timeleft < 0 )
		{
			if( i < Bullets.Size() - 1 )
				Bullets[ i ] = Bullets.Last();
			Bullets.Erase( Bullets.Size() - 1 );
			i--;
		}
	}

	// Process the collectibles
	for( UINT32 i = 0; i < Collectibles.Size(); ++i )
	{
		if( Collectibles[ i ].TimeState < 1 )
		{
			if( ( Collectibles[ i ].TimeState -= delta ) < 0 )
			{
				if( i < Collectibles.Size() - 1 )
					Collectibles[ i ] = Collectibles.Last();
				Collectibles.Erase( Collectibles.Size() - 1 );
			}
		}
		else if( Player.LifeState == 0 && ( Collectibles[ i ].Pos - Player.Pos ).LengthSq() <= PLAYER_RADIUS + COLLECT_COL_RADIUS )
		{
			Collectibles[ i ].TimeState = 0.99999f;
			SndOrbHit.Stop();
			SndOrbHit.Play();
		}
	}

	Timeout -= delta;

	// Process the physics
	// - step 1: collect all bodies
	TArray< SBPhyBody* > Bodies;
	Bodies.Reserve( Enemies.Size() + 1 );
	Bodies.PushBack( &Player );
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		if( Enemies[ i ].LifeState <= 0 )
			Bodies.PushBack( &Enemies[ i ] );
	}

	// - step 2: physics
	const FLOAT rest = 0.55f;
	const FLOAT ifrc = 0.99f;
	const FLOAT move = 1.0001f;
	for( UINT32 it = 0; it < 1; ++it )
	{
		// 2.1: dyn-dyn
		for( UINT32 i = 0; i < Bodies.Size(); ++i )
		{
			for( UINT32 j = i + 1; j < Bodies.Size(); ++j )
			{
				SBPhyBody* b1 = Bodies[ i ];
				SBPhyBody* b2 = Bodies[ j ];

				FLOAT r1 = PLAYER_RADIUS;	// INCORRECT
				FLOAT r2 = ENEMY_RADIUS;	// INCORRECT
				FLOAT lsq = ( b1->Pos - b2->Pos ).LengthSq();

				if( lsq < r1 * r1 + r2 * r2 )
				{
					FLOAT exh = ( r1 + r2 - gSqrt( lsq ) ) * 0.5f;
					mVec2 n = lsq ? ( b2->Pos - b1->Pos ).Normalized() : mVec2( 0.0f );
					mVec2 t = n.Perp();

					b1->Pos -= n * exh * move;
					b2->Pos += n * exh * move;

					FLOAT nv1 = b1->Vel * n;
					FLOAT tv1 = b1->Vel * t;
					FLOAT nv2 = b2->Vel * n;
					FLOAT tv2 = b2->Vel * t;
					nv1 *= -rest;
					tv1 *= ifrc;
					nv2 *= -rest;
					tv2 *= ifrc;
					b1->Vel = n * nv1 + t * tv1;
					b2->Vel = n * nv2 + t * tv2;

					b1->CrashAmount -= nv1;
					b2->CrashAmount += nv2;
				}
			}
		}
		// 2.2: dyn-stat
		for( UINT32 i = 0; i < Bodies.Size(); ++i )
		{
			SBPhyBody& B = *Bodies[ i ];
			FLOAT R = PLAYER_RADIUS;	// INCORRECT
			INT32 minx = gFloor( B.Pos.x - R );
			INT32 miny = gFloor( B.Pos.y - R );
			INT32 maxx = gCeil( B.Pos.x + R );
			INT32 maxy = gCeil( B.Pos.y + R );
			minx = MAX( 0, MIN( Width - 1, minx ) );
			maxx = MAX( 0, MIN( Width - 1, maxx ) );
			miny = MAX( 0, MIN( Height - 1, miny ) );
			maxy = MAX( 0, MIN( Height - 1, maxy ) );

			for( INT32 y = miny; y <= maxy; ++y )
			{
				for( INT32 x = minx; x <= maxx; ++x )
				{
					if( !Walls[ x + Width * y ] )
						continue;

					mAABB2 box( x - 0.5f, y - 0.5f, x + 0.5f, y + 0.5f );

					mVec2& ppos = B.Pos;
					mVec2 dist = mVec2( 0.0f );
					if( ppos.x < box.lo.x ) dist.x = ppos.x - box.lo.x; else if( ppos.x > box.hi.x ) dist.x = ppos.x - box.hi.x;
					if( ppos.y < box.lo.y ) dist.y = ppos.y - box.lo.y; else if( ppos.y > box.hi.y ) dist.y = ppos.y - box.hi.y;
					dist = -dist;

					if( dist.LengthSq() < R * R )
					{
						// process the collision
						// - find the normal / tangent
						bool inside = false;
						mVec2 nrm( 0.0f );
						if( dist != nrm ) nrm = -dist.Normalized();
						else if( B.Pos - mVec2( x, y ) != nrm )
						{
							if( gAbs( B.Pos.x - x ) > gAbs( B.Pos.y - y ) ) nrm.Set( B.Pos.x > x, 0 );
							else nrm.Set( 0, B.Pos.y > 0 );
							inside = true;
						}
						mVec2 tng = nrm.Perp();

						// - process velocity
						FLOAT nv = B.Vel * nrm;
						FLOAT tv = B.Vel * tng;
						nv *= -rest;
						tv *= ifrc;
						B.Vel = nv * nrm + tv * tng;
						B.CrashAmount += nv * 0.1f;

						// - process position
						FLOAT pendist = inside ? ( 0.5f + R - nrm * ( B.Pos - mVec2( x, y ) ) ) : R - dist.Length();
						B.Pos += pendist * nrm * move;
					}
				}
			}
		}
	}

	// Render
	Process( delta );
}

/// - - - -
//
//   R E N D E R I N G
//
/// - - - -

void SBLevel::Do()
{
	InitShaders();

	// prepare
	mMatrix mtx;
	FLOAT aspect = FLOAT( IO->GetWidth() ) / FLOAT( IO->GetHeight() );
	M_MatProjOrthoOffCenter( mtx, CamPos.x - CamSize * aspect, CamPos.x + CamSize * aspect, CamPos.y - CamSize, CamPos.y + CamSize, -1, 1 );

	mVec2 ppos = Player.Pos;
	FLOAT pang = Player.Angle;
	mVec2 pdir( gSin( pang ), gCos( pang ) );

	// RENDER

	FB_Begin( FBRB );	// >> LIGHTING

	// Render the lighting
	FB_Begin( FBLighting );
	// FB_Clear( mFColor( 0.1f, 0.15f, 0.3f ) );
	FB_Clear( mFColor( 0.11f, 0.25f, 0.5f ) );

	bool hascol = false;
	for( UINT32 i = 0; i < Collectibles.Size() && !hascol; ++i )
		if( Collectibles[ i ].TimeState < 1.0f && Collectibles[ i ].TimeState > 0.95f )
			hascol = true;

	if( Player.Exists && hascol )
	{
		mVec2 pdn = mVec2( gSin( pang + PI ), gCos( pang + PI ) ) * PLAYER_RADIUS * 1.2f;
		mVec2 prt = mVec2( gSin( pang + PI * 0.5f ), gCos( pang + PI * 0.5f ) ) * PLAYER_RADIUS * 1.2f;
		M_SetShading( MBM_ALPHA, SH_Get( "Shaders/Player.shd" ) );
		M_SetTexture( 0, TexPlayer );
		SUIVtx vts[ 4 ] = 
		{
			{ ppos - prt - pdn, mVec2( 0, 0 ), 0.0f },
			{ ppos + prt - pdn, mVec2( 1, 0 ), 0.0f },
			{ ppos + prt + pdn, mVec2( 1, 1 ), 0.0f },
			{ ppos - prt + pdn, mVec2( 0, 1 ), 0.0f },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}


	// point lights
	M_SetShading( MBM_ADD, SH_Get( "Shaders/PointLight.shd" ) );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	for( UINT32 i = 0; i < Lights.Size(); ++i )
	{
		mVec2 pos = Lights[ i ];
		FLOAT rad = 3.14159f * 3;
		if( !InCamera( pos, rad ) )
			continue;

		SUIVtx vts[] = 
		{
			{ pos + mVec2( -rad, -rad ), mVec2( -1.0f, -1.0f ), 0.0f },
			{ pos + mVec2( rad, -rad ), mVec2( 1.0f, -1.0f ), 0.0f },
			{ pos + mVec2( rad, rad ), mVec2( 1.0f, 1.0f ), 0.0f },
			{ pos + mVec2( -rad, rad ), mVec2( -1.0f, 1.0f ), 0.0f },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}

	// --- B U L L E T S
	// - flash
	M_SetShading( MBM_ADD, SH_Get( "Shaders/Flash.shd" ) );
	M_SetTexture( 0, TexFlash );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	for( UINT32 i = 0; i < Bullets.Size(); ++i )
	{
		mVec2 pos = Bullets[ i ].Pos + ( mVec2( LayeredSineNoise( Bullets[ i ].Random * 3.2523f ), 
			LayeredSineNoise( Bullets[ i ].Random + Bullets[ i ].Timeleft ) ) * 2.0f - mVec2( 1.0f ) ) * 0.05f;
		FLOAT q = 0.4f + Bullets[ i ].Random * 0.1f;
		mVec2 dn = -Bullets[ i ].Dir * q;
		mVec2 rt = Bullets[ i ].Dir.Perp2() * q;
		mFColor col( 4.1f, 4.0f, 3.9f, Bullets[ i ].Timeleft / Bullets[ i ].Lifetime );
		SUIVtx vts[] = 
		{
			{ pos - rt - dn, mVec2( 0, 0 ), col },
			{ pos + rt - dn, mVec2( 1, 0 ), col },
			{ pos + rt + dn, mVec2( 1, 1 ), col },
			{ pos - rt + dn, mVec2( 0, 1 ), col },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );

	}
	// - line
	M_SetShading( MBM_ADD, SH_Get( "Shaders/Color.shd" ) );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	for( UINT32 i = 0; i < Bullets.Size(); ++i )
	{
		mVec2 pos = Bullets[ i ].Pos;
		mVec2 dir = Bullets[ i ].Dir * Bullets[ i ].RayDist;
		mFColor col( 2.1f, 1.5f, 0.5f, Bullets[ i ].Timeleft / Bullets[ i ].Lifetime );
		mFColor cola = col; cola.a *= 1.0f - Bullets[ i ].RayDist / Bullets[ i ].Distance;
		SUIVtx vts[] =
		{
			{ pos, 0.0f, col },
			{ pos + dir, 0.0f, cola },
		};
		DrawUIVtx( vts, 2, PT_Lines );
	}
	// - hit
	for( UINT32 i = 0; i < Bullets.Size(); ++i )
	{
		if( Bullets[ i ].RayDist < Bullets[ i ].Distance )
			Bullets[ i ].HitFX.Render( mtx );
	}
	
	// --- L A S E R  of the player
	if( Player.Exists && Player.ShootTimeout <= 0 )
	{
		FLOAT dist = Player.Ammo ? 10.0f : 1.0f;
		FLOAT raydist = RaycastWalls( ppos, pdir, dist );
		RaycastEnts( ppos, pdir, raydist, Enemies.Size() );
		mFColor lcol( 2.1f, 0.5f, 0.1f, 0.5f * ( 1 - gAbs( Player.LifeState ) ) );
		mFColor lcola = lcol; lcola.a *= 1.0f - raydist / dist;
		SUIVtx vts[] =
		{
			{ ppos + pdir * 0.125f, 0.0f, lcol },
			{ ppos + pdir * raydist, 0.0f, lcola },
		};
		DrawUIVtx( vts, 2, PT_Lines );
	}
	
	// --- C O L L E C T I B L E S  emissive
	M_SetShading( MBM_ADDPREMUL, SH_Get( "Shaders/Collectible.shd" ) );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	M_SetTexture( 0, TexCollect );
	for( UINT32 i = 0; i < Collectibles.Size(); ++i )
	{
		mAABB2 box( Collectibles[ i ].Pos - mVec2( COLLECT_RADIUS ), Collectibles[ i ].Pos + mVec2( COLLECT_RADIUS ) );
		mFColor cc = Collectibles[ i ].TimeState;
		SUIVtx vts[ 4 ] = 
		{
			{ box.v_xy(), mVec2( 0, 0 ), cc },
			{ box.v_Xy(), mVec2( 1, 0 ), cc },
			{ box.v_XY(), mVec2( 1, 1 ), cc },
			{ box.v_xY(), mVec2( 0, 1 ), cc },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}

	// --- enemy  L I G H T S
	M_SetTexture( 0, TexEnemyLight );
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		if( Enemies[ i ].LifeState != 0 )
			continue;

		mVec2 pos = Enemies[ i ].Pos;
		mVec2 dn( gSin( Enemies[ i ].Angle ), gCos( Enemies[ i ].Angle ) );
		mVec2 rt = dn.Perp();
		FLOAT size = 2;
		dn *= size;
		rt *= size;
		pos += dn;
		mFColor col = Lerp( mFColor( 3.1f, 0.5f, 0.2f ), mFColor( 0.9f, 0.8f, 0.7f ), FLOAT( Enemies[ i ].AlertState <= 0 ) );
		SUIVtx vts[] = 
		{
			{ pos - rt - dn, mVec2( 0, 0 ), col },
			{ pos + rt - dn, mVec2( 1, 0 ), col },
			{ pos + rt + dn, mVec2( 1, 1 ), col },
			{ pos - rt + dn, mVec2( 0, 1 ), col },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}
	
	// --- enemy  E X P L O S I O N S
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		Enemies[ i ].DieFX.Render( mtx );
	}

	FB_End( FBLighting );	// << LIGHTING

	// < in FBRB
	FB_Clear( 0.0f );

	// --- W A L L S
	M_SetShading( 0, SH_Get( "Shaders/LevelTiles.shd" ) );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	M_SetTexture( 2, TexLightmap );
	for( UINT32 i = 0; i < Walls.Size(); ++i )
	{
		mVec2 pos = mVec2( i % Width, i / Width );
		mAABB2 box( pos - mVec2( 0.5f ), pos + mVec2( 0.5f ) );
		if( !InCamera( box ) )
			continue;

		if( Walls[ i ] )
		{
			M_SetTexture( 0, TexWall );
			M_SetTexture( 1, TexWallOvr[ Adjacency[ i ] ] );
		}
		else
		{
			M_SetTexture( 0, TexFloor );
			M_SetTexture( 1, TexNull );
		}
		SUIVtx vts[ 4 ] = 
		{
			{ box.v_xy(), mVec2( 0, 0 ), mVec4( box.lo.x + 0.5f, box.lo.y + 0.5f, 0, 0 ) },
			{ box.v_Xy(), mVec2( 1, 0 ), mVec4( box.hi.x + 0.5f, box.lo.y + 0.5f, 0, 0 ) },
			{ box.v_XY(), mVec2( 1, 1 ), mVec4( box.hi.x + 0.5f, box.hi.y + 0.5f, 0, 0 ) },
			{ box.v_xY(), mVec2( 0, 1 ), mVec4( box.lo.x + 0.5f, box.hi.y + 0.5f, 0, 0 ) },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}
	
	// --- A C T O R S
	M_SetShading( MBM_ALPHA, SH_Get( "Shaders/Player.shd" ) );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	// The player..
	if( Player.Exists )
	{
		mVec2 pdn = mVec2( gSin( pang + PI ), gCos( pang + PI ) ) * PLAYER_RADIUS;
		mVec2 prt = mVec2( gSin( pang + PI * 0.5f ), gCos( pang + PI * 0.5f ) ) * PLAYER_RADIUS;
		M_SetTexture( 0, TexPlayer );
		FLOAT col = Player.LifeState + 1.0f;
		SUIVtx vts[ 4 ] = 
		{
			{ ppos - prt - pdn, mVec2( 0, 0 ), col },
			{ ppos + prt - pdn, mVec2( 1, 0 ), col },
			{ ppos + prt + pdn, mVec2( 1, 1 ), col },
			{ ppos - prt + pdn, mVec2( 0, 1 ), col },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );

		Player.FxSparks.Render( mtx );
	}
	// The enemies..
	if( Enemies.Size() )
	{
		M_SetShading( MBM_ALPHA, SH_Get( "Shaders/Player.shd" ) );
		M_SetTexture( 0, TexEnemy );
	}
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		mVec2 epos = Enemies[ i ].Pos;
		FLOAT eang = Enemies[ i ].Angle;
		mVec2 pdn = mVec2( gSin( eang + PI ), gCos( eang + PI ) ) * ENEMY_RADIUS;
		mVec2 prt = mVec2( gSin( eang + PI * 0.5f ), gCos( eang + PI * 0.5f ) ) * ENEMY_RADIUS;
		FLOAT col = MAX( -1, MIN( 1, Enemies[ i ].LifeState ) ) + 1.0f;
		SUIVtx vts[ 4 ] = 
		{
			{ epos - prt - pdn, mVec2( 0, 0 ), col },
			{ epos + prt - pdn, mVec2( 1, 0 ), col },
			{ epos + prt + pdn, mVec2( 1, 1 ), col },
			{ epos - prt + pdn, mVec2( 0, 1 ), col },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}

	// --- C O L L E C T I B L E S  - diffuse
	M_SetShading( MBM_ALPHAPREMUL, SH_Get( "Shaders/Collectible.shd" ) );
	M_GetCurShader()->GetVariable( "SG_mProj" )->SetF( mtx.fa, 4, TRUE );
	M_SetTexture( 0, TexCollect );
	M_SetTexture( 1, TexColNoise );
	for( UINT32 i = 0; i < Collectibles.Size(); ++i )
	{
		mAABB2 box( Collectibles[ i ].Pos - mVec2( COLLECT_RADIUS ), Collectibles[ i ].Pos + mVec2( COLLECT_RADIUS ) );
		mAABB2 tb = PreTransformBox( box );
		mFColor cc = Collectibles[ i ].TimeState;
		SUIVtx vts[ 4 ] = 
		{
			{ tb.v_xy(), mVec2( 0, 0 ), cc },
			{ tb.v_Xy(), mVec2( 1, 0 ), cc },
			{ tb.v_XY(), mVec2( 1, 1 ), cc },
			{ tb.v_xY(), mVec2( 0, 1 ), cc },
		};
		DrawUIVtx( vts, 4, PT_TriangleFan );
	}

	// --- E X P L O S I O N S
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		Enemies[ i ].DieFX.Render( mtx );
	}

	// --- L I G H T I N G
	M_SetShading( MBMF_BLEND( MBMF_BLEND_ADD, MBMF_BF_DC, MBMF_BF_ZERO ), SH_Get( "Shaders/FSQuad.shd" ) );
	M_SetTexture( 0, TexLighting );
	FSBlit();

	//
	// U S E R   I N T E R F A C E
	//

	if( DrawUI )
	{

		// Draw AMMO
		mAABB2 ammorect1( 64, 64, 512, 512 );
		mAABB2 ammorect2( 160, 64, 512, 512 );
		CurMaterial = &MatTexture;
		DrawTextRect( "AMMO:", 5, mFColor( 0.7f, 0.75f, 0.9f ), 0, ammorect1, 32.0f );
		TString ammolist1;
		UINT32 acur = Player.Ammo % AMMO_DIV;
		if( acur == 0 && Player.Ammo ) acur = AMMO_DIV;
		mFColor acol( 0.1f, 0.5f, 0.05f ), acol1( 0.5f, 0.4f, 0.1f ), acol2( 0.7f, 0.1f, 0.05f );
		mFColor acols[ 10 ] = { acol2, acol2, acol1, acol1, acol, acol, acol, acol, acol, acol };
		const TCHAR* AMMOSYM = "I";
		const UINT32 AMMOSYMLEN = 1;
		const FLOAT AMMOSYMMUL = 0.8f;
		for( UINT32 i = 1; i <= AMMO_DIV; ++i )
		{
			mFColor col = i > acur ? mFColor( 0.0f, 0.5f ) : acols[ i - 1 ];
			DrawTextRect( AMMOSYM, AMMOSYMLEN, mFColor( 0.0f, 0.5f ), 0, mAABB2( ammorect2.lo + mVec2( 4.0f ), ammorect2.hi ), 32.0f );
			DrawTextRect( AMMOSYM, AMMOSYMLEN, col, 0, ammorect2, 32.0f );
			ammorect2.lo.x += CurFont->GetTextLength( 32.0f, AMMOSYM, 0, AMMOSYMLEN ) * AMMOSYMMUL;
		}
		ammorect2.lo.Set( 160, 96 );
		INT32 ammo1 = Player.Ammo - 1;
		UINT32 amax = MAX( ammo1, 0 ) / AMMO_DIV;
		for( UINT32 i = 0; i < amax; ++i )
		{
			mFColor col = mFColor( 0.5f, 0.5f, 0.5f );
			DrawTextRect( AMMOSYM, AMMOSYMLEN, mFColor( 0.0f, 0.5f ), 0, mAABB2( ammorect2.lo + mVec2( 4.0f ), ammorect2.hi ), 32.0f );
			DrawTextRect( AMMOSYM, AMMOSYMLEN, col, 0, ammorect2, 32.0f );
			ammorect2.lo.x += CurFont->GetTextLength( 32.0f, AMMOSYM, 0, AMMOSYMLEN ) * AMMOSYMMUL;
		}

		// Draw central info (time left, collectible count) and right-side info (health)
		INT32 itc = 0;
		for( UINT32 i = 0; i < Collectibles.Size(); ++i )
			if( Collectibles[ i ].TimeState == 1.0f )
				itc++;
	
		mVec2 safebor( 64.0f );
		INT32 safetime = gCeil( Timeout );
		safetime = MAX( safetime, 0 );
		INT32 stmins = safetime / 60;
		INT32 stsecs = safetime % 60;
		INT32 safehealth = MAX( 0, Player.Health );
		TString tm, tm2;
		const TCHAR* prestr = "", * poststr = "";
		if( safetime <= 20 )
		{
			if( safetime % 2 == 0 )	prestr = " ";
			else					poststr = " ";
		}
		tm.Printf( 128, "%sTime left: %d:%02d%s\nOrbs left: %d", prestr, stmins, stsecs, poststr, itc );
		tm2.Printf( 128, "HEALTH\n[%d]", safehealth );
		DrawTextRect( tm.CStr(), tm.Size(), mFColor( 0.9f, 0.7f ), TF_CENTER, mAABB2( Rect.lo + safebor, Rect.hi - safebor ), 32.0f );
		DrawTextRect( tm2.CStr(), tm2.Size(), safehealth < 5 ? mFColor( 0.9f, 0.2f, 0.1f, 0.7f ) : mFColor( 0.9f, 0.7f ), TF_RIGHT, mAABB2( Rect.lo + safebor, Rect.hi - safebor ), 32.0f );

	}

	//
	// U . I .   E N D
	//

	FB_End( FBRB );
	M_SetShading( MBM_ALPHA, SH_Get( "Shaders/PostProcess.shd" ) );
	M_SetTexture( 0, TexRB );
	FSBlit();
}


mAABB2 SBLevel::GetCamAABB()
{
	FLOAT aspect = FLOAT( IO->GetWidth() ) / FLOAT( IO->GetHeight() );
	return mAABB2( CamPos.x - CamSize * aspect, CamPos.y - CamSize, CamPos.x + CamSize * aspect, CamPos.y + CamSize );
}

UBOOL SBLevel::InCamera( const mVec2& pos, FLOAT rad )
{
	mVec2 dst = GetCamAABB().GetDistance( pos );
	return dst.LengthSq() < rad * rad;
}

UBOOL SBLevel::InCamera( const mAABB2& box )
{
	return GetCamAABB().Intersect( box );
}


FLOAT mod01( FLOAT x )
{
	FLOAT m = gMod( x, 1 );
	if( m < 0 ) m += 1;
	return m;
}
FLOAT SBLevel::RaycastWalls( mVec2 pos, mVec2 dir, FLOAT maxdist )
{
	// if point outside world, do nothing
	INT32 vX = gRound( pos.x );
	INT32 vY = gRound( pos.y );
	if( vX < 0 || vX >= Width ||
		vY < 0 || vY >= Height )
		return maxdist;

	if( Walls[ vX + Width * vY ] )
		return 0;

	INT32 stepX = dir.x > 0 ? 1 : -1;
	INT32 stepY = dir.y > 0 ? 1 : -1;
	FLOAT tDeltaX = dir.x == 0 ? FLT_MAX : FLOAT( stepX ) / dir.x;
	FLOAT tDeltaY = dir.y == 0 ? FLT_MAX : FLOAT( stepY ) / dir.y;
	FLOAT tMaxX = tDeltaX * ( FLOAT( dir.x > 0 ) - stepX * mod01( pos.x + 0.5f ) );
	FLOAT tMaxY = tDeltaY * ( FLOAT( dir.y > 0 ) - stepY * mod01( pos.y + 0.5f ) );

	for(;;)
	{
		if( tMaxX < tMaxY )
		{
			vX += stepX;
			if( Walls[ vX + vY * Width ] )
				return tMaxX;
			tMaxX += tDeltaX;
			if( vX < 0 || vX >= Width )
				break;
		}
		else
		{
			vY += stepY;
			if( Walls[ vX + vY * Width ] )
				return tMaxY;
			tMaxY += tDeltaY;
			if( vY < 0 || vY >= Height )
				break;
		}
		if( MIN( tMaxX, tMaxY ) > maxdist )
			break;
	}
	return maxdist;
}


UBOOL rayCircle( mVec2 pos, mVec2 dir, mVec2 cpos, FLOAT crad, FLOAT& maxdist )
{
	mRay3 r;
	r.vPos.Set( pos.x, pos.y, 0 );
	r.vDir.Set( dir.x, dir.y, 0 );
	FLOAT dst;
	if( r.IntersectSphere( mVec3( cpos ), crad, &dst ) && dst < maxdist )
	{
		maxdist = dst;
		return TRUE;
	}
	return FALSE;
}

INT32 SBLevel::RaycastEnts( mVec2 pos, mVec2 dir, FLOAT& maxdist, INT32 ignore )
{
	INT32 out = -1;
	for( UINT32 i = 0; i < Enemies.Size(); ++i )
	{
		if( i == ignore || Enemies[ i ].LifeState != 0 )
			continue;

		// ray-circle test
		if( rayCircle( pos, dir, Enemies[ i ].Pos, ENEMY_RADIUS, maxdist ) )
			out = i;
	}
	if( Enemies.Size() != ignore )
	{
		if( Player.LifeState == 0 && rayCircle( pos, dir, Player.Pos, PLAYER_RADIUS, maxdist ) )
			out = Enemies.Size();
	}
	return out;
}


mVec2 bGenNormal( const mVec2& pos, const mVec2& dir )
{
	if( gAbs( gRound( pos.x ) - pos.x ) < gAbs( gRound( pos.y ) - pos.y ) )
	{
		// on X, output Y
		return mVec2( 0, dir.y > 0 ? -1 : 1 );
	}
	else
	{
		// on Y, output X
		return mVec2( dir.x > 0 ? -1 : 1, 0 );
	}
}

void SBLevel::Shoot( mVec2 pos, mVec2 dir, FLOAT dist, FLOAT time, FLOAT rand, INT32 from )
{
	pos += dir * 0.125f;
	mVec2 bdir = dir.Rotated( ( M_RandF() * PI - PI * 0.5f ) * rand );

	FLOAT raydist = RaycastWalls( pos, bdir, dist );
	INT32 ent = RaycastEnts( pos, bdir, raydist, from );
	if( ent >= 0 )
	{
		// apply megahurtz
		if( ent == Enemies.Size() ) // player
		{
			Player.CrashAmount += 0.1f;
			Player.Health -= 1;
			if( Player.Health <= 0 )
			{
				Player.LifeState += 0.001f;
			}
		}
		else // enemy
		{
			SBEnemy& E = Enemies[ ent ];
			E.CrashAmount += 0.2f;
			E.Health -= 1;
			if( E.Health <= 0 && E.LifeState <= 0 )
			{
				E.LifeState += 0.001f;
				E.DieFX.Trigger( E.Pos, 1.0f );
			}
		}
	}

	SBBullet blt = { pos, bdir, time, time, dist, raydist, from == Enemies.Size(), M_RandF(), SBSparkFX() };
	if( raydist < dist )
	{
		blt.HitFX.Trigger( pos + bdir * raydist, bGenNormal( pos, bdir ), 0.3f );
	}
	Bullets.PushBack( blt );
}

